@@ -22,6 +22,9 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 // KWin
 #include "../../src/client/compositor.h"
 #include "../../src/client/connection_thread.h"
+#include "../../src/client/datadevice.h"
+#include "../../src/client/datadevicemanager.h"
+#include "../../src/client/datasource.h"
 #include "../../src/client/event_queue.h"
 #include "../../src/client/keyboard.h"
 #include "../../src/client/pointer.h"
@@ -31,7 +34,10 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "../../src/client/shm_pool.h"
 #include "../../src/server/buffer_interface.h"
 #include "../../src/server/compositor_interface.h"
+#include "../../src/server/datadevicemanager_interface.h"
 #include "../../src/server/display.h"
+#include "../../src/server/keyboard_interface.h"
+#include "../../src/server/pointer_interface.h"
 #include "../../src/server/seat_interface.h"
 #include "../../src/server/surface_interface.h"
 // Wayland
@@ -52,9 +58,12 @@ private Q_SLOTS:
     void testCapabilities_data();
     void testCapabilities();
     void testPointer();
+    void testPointerButton_data();
+    void testPointerButton();
     void testKeyboard();
     void testCast();
     void testDestroy();
+    void testSelection();
     // TODO: add test for keymap
 
 private:
@@ -278,15 +287,24 @@ void TestWaylandSeat::testPointer()
     SurfaceInterface *serverSurface = surfaceCreatedSpy.first().first().value<KWayland::Server::SurfaceInterface*>();
     QVERIFY(serverSurface);
 
-    PointerInterface *serverPointer = m_seatInterface->pointer();
-    serverPointer->setGlobalPos(QPoint(20, 18));
-    serverPointer->setFocusedSurface(serverSurface, QPoint(10, 15));
-    // no pointer yet - won't be set
-    QVERIFY(!serverPointer->focusedSurface());
+    m_seatInterface->setPointerPos(QPoint(20, 18));
+    m_seatInterface->setFocusedPointerSurface(serverSurface, QPoint(10, 15));
+    // no pointer yet
+    QVERIFY(m_seatInterface->focusedPointerSurface());
+    QVERIFY(!m_seatInterface->focusedPointer());
 
     Pointer *p = m_seat->createPointer(m_seat);
+    const Pointer &cp = *p;
     QVERIFY(p->isValid());
-    wl_display_flush(m_connection->display());
+    QSignalSpy pointerCreatedSpy(m_seatInterface, SIGNAL(pointerCreated(KWayland::Server::PointerInterface*)));
+    QVERIFY(pointerCreatedSpy.isValid());
+    // once the pointer is created it should be set as the focused pointer
+    QVERIFY(pointerCreatedSpy.wait());
+    QVERIFY(m_seatInterface->focusedPointer());
+    QCOMPARE(pointerCreatedSpy.first().first().value<PointerInterface*>(), m_seatInterface->focusedPointer());
+
+    m_seatInterface->setFocusedPointerSurface(nullptr);
+    serverSurface->client()->flush();
     QTest::qWait(100);
 
     QSignalSpy enteredSpy(p, SIGNAL(entered(quint32,QPointF)));
@@ -304,25 +322,31 @@ void TestWaylandSeat::testPointer()
     QSignalSpy buttonSpy(p, SIGNAL(buttonStateChanged(quint32,quint32,quint32,KWayland::Client::Pointer::ButtonState)));
     QVERIFY(buttonSpy.isValid());
 
-    serverPointer->setFocusedSurface(serverSurface, QPoint(10, 15));
-    QCOMPARE(serverPointer->focusedSurface(), serverSurface);
+    QVERIFY(!p->enteredSurface());
+    QVERIFY(!cp.enteredSurface());
+    m_seatInterface->setFocusedPointerSurface(serverSurface, QPoint(10, 15));
+    QCOMPARE(m_seatInterface->focusedPointerSurface(), serverSurface);
     QVERIFY(enteredSpy.wait());
     QCOMPARE(enteredSpy.first().first().value<quint32>(), m_display->serial());
     QCOMPARE(enteredSpy.first().last().toPoint(), QPoint(10, 3));
+    PointerInterface *serverPointer = m_seatInterface->focusedPointer();
+    QVERIFY(serverPointer);
+    QCOMPARE(p->enteredSurface(), s);
+    QCOMPARE(cp.enteredSurface(), s);
 
     // test motion
-    serverPointer->updateTimestamp(1);
-    serverPointer->setGlobalPos(QPoint(10, 16));
+    m_seatInterface->setTimestamp(1);
+    m_seatInterface->setPointerPos(QPoint(10, 16));
     QVERIFY(motionSpy.wait());
     QCOMPARE(motionSpy.first().first().toPoint(), QPoint(0, 1));
     QCOMPARE(motionSpy.first().last().value<quint32>(), quint32(1));
 
     // test axis
-    serverPointer->updateTimestamp(2);
-    serverPointer->axis(Qt::Horizontal, 10);
+    m_seatInterface->setTimestamp(2);
+    m_seatInterface->pointerAxis(Qt::Horizontal, 10);
     QVERIFY(axisSpy.wait());
-    serverPointer->updateTimestamp(3);
-    serverPointer->axis(Qt::Vertical, 20);
+    m_seatInterface->setTimestamp(3);
+    m_seatInterface->pointerAxis(Qt::Vertical, 20);
     QVERIFY(axisSpy.wait());
     QCOMPARE(axisSpy.first().at(0).value<quint32>(), quint32(2));
     QCOMPARE(axisSpy.first().at(1).value<Pointer::Axis>(), Pointer::Axis::Horizontal);
@@ -333,20 +357,20 @@ void TestWaylandSeat::testPointer()
     QCOMPARE(axisSpy.last().at(2).value<qreal>(), qreal(20));
 
     // test button
-    serverPointer->updateTimestamp(4);
-    serverPointer->buttonPressed(1);
+    m_seatInterface->setTimestamp(4);
+    m_seatInterface->pointerButtonPressed(1);
     QVERIFY(buttonSpy.wait());
     QCOMPARE(buttonSpy.at(0).at(0).value<quint32>(), m_display->serial());
-    serverPointer->updateTimestamp(5);
-    serverPointer->buttonPressed(2);
+    m_seatInterface->setTimestamp(5);
+    m_seatInterface->pointerButtonPressed(2);
     QVERIFY(buttonSpy.wait());
     QCOMPARE(buttonSpy.at(1).at(0).value<quint32>(), m_display->serial());
-    serverPointer->updateTimestamp(6);
-    serverPointer->buttonReleased(2);
+    m_seatInterface->setTimestamp(6);
+    m_seatInterface->pointerButtonReleased(2);
     QVERIFY(buttonSpy.wait());
     QCOMPARE(buttonSpy.at(2).at(0).value<quint32>(), m_display->serial());
-    serverPointer->updateTimestamp(7);
-    serverPointer->buttonReleased(1);
+    m_seatInterface->setTimestamp(7);
+    m_seatInterface->pointerButtonReleased(1);
     QVERIFY(buttonSpy.wait());
     QCOMPARE(buttonSpy.count(), 4);
 
@@ -362,14 +386,14 @@ void TestWaylandSeat::testPointer()
     QCOMPARE(buttonSpy.at(1).at(2).value<quint32>(), quint32(2));
     QCOMPARE(buttonSpy.at(1).at(3).value<KWayland::Client::Pointer::ButtonState>(), KWayland::Client::Pointer::ButtonState::Pressed);
 
-    QCOMPARE(buttonSpy.at(2).at(0).value<quint32>(), serverPointer->buttonSerial(2));
+    QCOMPARE(buttonSpy.at(2).at(0).value<quint32>(), m_seatInterface->pointerButtonSerial(2));
     // timestamp
     QCOMPARE(buttonSpy.at(2).at(1).value<quint32>(), quint32(6));
     // button
     QCOMPARE(buttonSpy.at(2).at(2).value<quint32>(), quint32(2));
     QCOMPARE(buttonSpy.at(2).at(3).value<KWayland::Client::Pointer::ButtonState>(), KWayland::Client::Pointer::ButtonState::Released);
 
-    QCOMPARE(buttonSpy.at(3).at(0).value<quint32>(), serverPointer->buttonSerial(1));
+    QCOMPARE(buttonSpy.at(3).at(0).value<quint32>(), m_seatInterface->pointerButtonSerial(1));
     // timestamp
     QCOMPARE(buttonSpy.at(3).at(1).value<quint32>(), quint32(7));
     // button
@@ -377,18 +401,117 @@ void TestWaylandSeat::testPointer()
     QCOMPARE(buttonSpy.at(3).at(3).value<KWayland::Client::Pointer::ButtonState>(), KWayland::Client::Pointer::ButtonState::Released);
 
     // leave the surface
-    serverPointer->setFocusedSurface(nullptr);
+    m_seatInterface->setFocusedPointerSurface(nullptr);
     QVERIFY(leftSpy.wait());
     QCOMPARE(leftSpy.first().first().value<quint32>(), m_display->serial());
+    QVERIFY(!p->enteredSurface());
+    QVERIFY(!cp.enteredSurface());
 
     // enter it again
-    serverPointer->setFocusedSurface(serverSurface, QPoint(0, 0));
+    m_seatInterface->setFocusedPointerSurface(serverSurface, QPoint(0, 0));
     QVERIFY(enteredSpy.wait());
+    QCOMPARE(p->enteredSurface(), s);
+    QCOMPARE(cp.enteredSurface(), s);
 
     delete s;
     wl_display_flush(m_connection->display());
     QTest::qWait(100);
-    QVERIFY(!serverPointer->focusedSurface());
+    QVERIFY(!m_seatInterface->focusedPointerSurface());
+}
+
+Q_DECLARE_METATYPE(Qt::MouseButton)
+
+void TestWaylandSeat::testPointerButton_data()
+{
+    QTest::addColumn<Qt::MouseButton>("qtButton");
+    QTest::addColumn<quint32>("waylandButton");
+
+    QTest::newRow("left")    << Qt::LeftButton    << quint32(BTN_LEFT);
+    QTest::newRow("right")   << Qt::RightButton   << quint32(BTN_RIGHT);
+    QTest::newRow("mid")     << Qt::MidButton     << quint32(BTN_MIDDLE);
+    QTest::newRow("middle")  << Qt::MiddleButton  << quint32(BTN_MIDDLE);
+    QTest::newRow("back")    << Qt::BackButton    << quint32(BTN_BACK);
+    QTest::newRow("x1")      << Qt::XButton1      << quint32(BTN_BACK);
+    QTest::newRow("extra1")  << Qt::ExtraButton1  << quint32(BTN_BACK);
+    QTest::newRow("forward") << Qt::ForwardButton << quint32(BTN_FORWARD);
+    QTest::newRow("x2")      << Qt::XButton2      << quint32(BTN_FORWARD);
+    QTest::newRow("extra2")  << Qt::ExtraButton2  << quint32(BTN_FORWARD);
+    QTest::newRow("task")    << Qt::TaskButton    << quint32(BTN_TASK);
+    QTest::newRow("extra3")  << Qt::ExtraButton3  << quint32(BTN_TASK);
+    QTest::newRow("extra4")  << Qt::ExtraButton4  << quint32(BTN_EXTRA);
+    QTest::newRow("extra5")  << Qt::ExtraButton5  << quint32(BTN_SIDE);
+    QTest::newRow("extra6")  << Qt::ExtraButton6  << quint32(0x118);
+    QTest::newRow("extra7")  << Qt::ExtraButton7  << quint32(0x119);
+    QTest::newRow("extra8")  << Qt::ExtraButton8  << quint32(0x11a);
+    QTest::newRow("extra9")  << Qt::ExtraButton9  << quint32(0x11b);
+    QTest::newRow("extra10") << Qt::ExtraButton10 << quint32(0x11c);
+    QTest::newRow("extra11") << Qt::ExtraButton11 << quint32(0x11d);
+    QTest::newRow("extra12") << Qt::ExtraButton12 << quint32(0x11e);
+    QTest::newRow("extra13") << Qt::ExtraButton13 << quint32(0x11f);
+}
+
+void TestWaylandSeat::testPointerButton()
+{
+    using namespace KWayland::Client;
+    using namespace KWayland::Server;
+
+    QSignalSpy pointerSpy(m_seat, SIGNAL(hasPointerChanged(bool)));
+    QVERIFY(pointerSpy.isValid());
+    m_seatInterface->setHasPointer(true);
+    QVERIFY(pointerSpy.wait());
+
+    QSignalSpy surfaceCreatedSpy(m_compositorInterface, SIGNAL(surfaceCreated(KWayland::Server::SurfaceInterface*)));
+    QVERIFY(surfaceCreatedSpy.isValid());
+    m_compositor->createSurface(m_compositor);
+    QVERIFY(surfaceCreatedSpy.wait());
+    SurfaceInterface *serverSurface = surfaceCreatedSpy.first().first().value<KWayland::Server::SurfaceInterface*>();
+    QVERIFY(serverSurface);
+
+    QScopedPointer<Pointer> p(m_seat->createPointer());
+    QVERIFY(p->isValid());
+    QSignalSpy buttonChangedSpy(p.data(), SIGNAL(buttonStateChanged(quint32,quint32,quint32,KWayland::Client::Pointer::ButtonState)));
+    QVERIFY(buttonChangedSpy.isValid());
+    wl_display_flush(m_connection->display());
+    QCoreApplication::processEvents();
+
+    m_seatInterface->setPointerPos(QPoint(20, 18));
+    m_seatInterface->setFocusedPointerSurface(serverSurface, QPoint(10, 15));
+    QVERIFY(m_seatInterface->focusedPointerSurface());
+    QVERIFY(m_seatInterface->focusedPointer());
+
+    QCoreApplication::processEvents();
+
+    m_seatInterface->setFocusedPointerSurface(serverSurface, QPoint(10, 15));
+    PointerInterface *serverPointer = m_seatInterface->focusedPointer();
+    QVERIFY(serverPointer);
+    QFETCH(Qt::MouseButton, qtButton);
+    QFETCH(quint32, waylandButton);
+    quint32 msec = QDateTime::currentMSecsSinceEpoch();
+    QCOMPARE(m_seatInterface->isPointerButtonPressed(waylandButton), false);
+    QCOMPARE(m_seatInterface->isPointerButtonPressed(qtButton), false);
+    m_seatInterface->setTimestamp(msec);
+    m_seatInterface->pointerButtonPressed(qtButton);
+    QCOMPARE(m_seatInterface->isPointerButtonPressed(waylandButton), true);
+    QCOMPARE(m_seatInterface->isPointerButtonPressed(qtButton), true);
+    QVERIFY(buttonChangedSpy.wait());
+    QCOMPARE(buttonChangedSpy.count(), 1);
+    QCOMPARE(buttonChangedSpy.last().at(0).value<quint32>(), m_seatInterface->pointerButtonSerial(waylandButton));
+    QCOMPARE(buttonChangedSpy.last().at(0).value<quint32>(), m_seatInterface->pointerButtonSerial(qtButton));
+    QCOMPARE(buttonChangedSpy.last().at(1).value<quint32>(), msec);
+    QCOMPARE(buttonChangedSpy.last().at(2).value<quint32>(), waylandButton);
+    QCOMPARE(buttonChangedSpy.last().at(3).value<KWayland::Client::Pointer::ButtonState>(), Pointer::ButtonState::Pressed);
+    msec = QDateTime::currentMSecsSinceEpoch();
+    m_seatInterface->setTimestamp(QDateTime::currentMSecsSinceEpoch());
+    m_seatInterface->pointerButtonReleased(qtButton);
+    QCOMPARE(m_seatInterface->isPointerButtonPressed(waylandButton), false);
+    QCOMPARE(m_seatInterface->isPointerButtonPressed(qtButton), false);
+    QVERIFY(buttonChangedSpy.wait());
+    QCOMPARE(buttonChangedSpy.count(), 2);
+    QCOMPARE(buttonChangedSpy.last().at(0).value<quint32>(), m_seatInterface->pointerButtonSerial(waylandButton));
+    QCOMPARE(buttonChangedSpy.last().at(0).value<quint32>(), m_seatInterface->pointerButtonSerial(qtButton));
+    QCOMPARE(buttonChangedSpy.last().at(1).value<quint32>(), msec);
+    QCOMPARE(buttonChangedSpy.last().at(2).value<quint32>(), waylandButton);
+    QCOMPARE(buttonChangedSpy.last().at(3).value<KWayland::Client::Pointer::ButtonState>(), Pointer::ButtonState::Released);
 }
 
 void TestWaylandSeat::testKeyboard()
@@ -409,30 +532,33 @@ void TestWaylandSeat::testKeyboard()
     SurfaceInterface *serverSurface = surfaceCreatedSpy.first().first().value<KWayland::Server::SurfaceInterface*>();
     QVERIFY(serverSurface);
 
-    KeyboardInterface *serverKeyboard = m_seatInterface->keyboard();
-
-    serverKeyboard->setFocusedSurface(serverSurface);
-    // no pointer yet - won't be set
-    QVERIFY(!serverKeyboard->focusedSurface());
+    m_seatInterface->setFocusedKeyboardSurface(serverSurface);
+    // no keyboard yet
+    QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
+    QVERIFY(!m_seatInterface->focusedKeyboard());
 
     Keyboard *keyboard = m_seat->createKeyboard(m_seat);
+    const Keyboard &ckeyboard = *keyboard;
     QVERIFY(keyboard->isValid());
     wl_display_flush(m_connection->display());
     QTest::qWait(100);
+    QVERIFY(m_seatInterface->focusedKeyboard());
 
-    serverKeyboard->updateTimestamp(1);
-    serverKeyboard->keyPressed(KEY_K);
-    serverKeyboard->updateTimestamp(2);
-    serverKeyboard->keyPressed(KEY_D);
-    serverKeyboard->updateTimestamp(3);
-    serverKeyboard->keyPressed(KEY_E);
+    m_seatInterface->setTimestamp(1);
+    m_seatInterface->keyPressed(KEY_K);
+    m_seatInterface->setTimestamp(2);
+    m_seatInterface->keyPressed(KEY_D);
+    m_seatInterface->setTimestamp(3);
+    m_seatInterface->keyPressed(KEY_E);
 
     QSignalSpy modifierSpy(keyboard, SIGNAL(modifiersChanged(quint32,quint32,quint32,quint32)));
     QVERIFY(modifierSpy.isValid());
 
-    // TODO: add a signalspy for enter
-    serverKeyboard->setFocusedSurface(serverSurface);
-    QCOMPARE(serverKeyboard->focusedSurface(), serverSurface);
+    QSignalSpy enteredSpy(keyboard, SIGNAL(entered(quint32)));
+    QVERIFY(enteredSpy.isValid());
+    m_seatInterface->setFocusedKeyboardSurface(serverSurface);
+    QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
+    QCOMPARE(m_seatInterface->focusedKeyboard()->focusedSurface(), serverSurface);
 
     // we get the modifiers sent after the enter
     QVERIFY(modifierSpy.wait());
@@ -441,24 +567,26 @@ void TestWaylandSeat::testKeyboard()
     QCOMPARE(modifierSpy.first().at(1).value<quint32>(), quint32(0));
     QCOMPARE(modifierSpy.first().at(2).value<quint32>(), quint32(0));
     QCOMPARE(modifierSpy.first().at(3).value<quint32>(), quint32(0));
+    QCOMPARE(enteredSpy.count(), 1);
+    QCOMPARE(enteredSpy.first().first().value<quint32>(), m_display->serial());
 
     QSignalSpy keyChangedSpy(keyboard, SIGNAL(keyChanged(quint32,KWayland::Client::Keyboard::KeyState,quint32)));
     QVERIFY(keyChangedSpy.isValid());
 
-    serverKeyboard->updateTimestamp(4);
-    serverKeyboard->keyReleased(KEY_E);
+    m_seatInterface->setTimestamp(4);
+    m_seatInterface->keyReleased(KEY_E);
     QVERIFY(keyChangedSpy.wait());
-    serverKeyboard->updateTimestamp(5);
-    serverKeyboard->keyReleased(KEY_D);
+    m_seatInterface->setTimestamp(5);
+    m_seatInterface->keyReleased(KEY_D);
     QVERIFY(keyChangedSpy.wait());
-    serverKeyboard->updateTimestamp(6);
-    serverKeyboard->keyReleased(KEY_K);
+    m_seatInterface->setTimestamp(6);
+    m_seatInterface->keyReleased(KEY_K);
     QVERIFY(keyChangedSpy.wait());
-    serverKeyboard->updateTimestamp(7);
-    serverKeyboard->keyPressed(KEY_F1);
+    m_seatInterface->setTimestamp(7);
+    m_seatInterface->keyPressed(KEY_F1);
     QVERIFY(keyChangedSpy.wait());
-    serverKeyboard->updateTimestamp(8);
-    serverKeyboard->keyReleased(KEY_F1);
+    m_seatInterface->setTimestamp(8);
+    m_seatInterface->keyReleased(KEY_F1);
     QVERIFY(keyChangedSpy.wait());
 
     QCOMPARE(keyChangedSpy.count(), 5);
@@ -478,7 +606,7 @@ void TestWaylandSeat::testKeyboard()
     QCOMPARE(keyChangedSpy.at(4).at(1).value<Keyboard::KeyState>(), Keyboard::KeyState::Released);
     QCOMPARE(keyChangedSpy.at(4).at(2).value<quint32>(), quint32(8));
 
-    serverKeyboard->updateModifiers(1, 2, 3, 4);
+    m_seatInterface->updateKeyboardModifiers(1, 2, 3, 4);
     QVERIFY(modifierSpy.wait());
     QCOMPARE(modifierSpy.count(), 2);
     QCOMPARE(modifierSpy.last().at(0).value<quint32>(), quint32(1));
@@ -486,19 +614,33 @@ void TestWaylandSeat::testKeyboard()
     QCOMPARE(modifierSpy.last().at(2).value<quint32>(), quint32(3));
     QCOMPARE(modifierSpy.last().at(3).value<quint32>(), quint32(4));
 
-    // TODO: add a test for leave signal
-    serverKeyboard->setFocusedSurface(nullptr);
-    QVERIFY(!serverKeyboard->focusedSurface());
+    QSignalSpy leftSpy(keyboard, SIGNAL(left(quint32)));
+    QVERIFY(leftSpy.isValid());
+    m_seatInterface->setFocusedKeyboardSurface(nullptr);
+    QVERIFY(!m_seatInterface->focusedKeyboardSurface());
+    QVERIFY(!m_seatInterface->focusedKeyboard());
+    QVERIFY(leftSpy.wait());
+    QCOMPARE(leftSpy.count(), 1);
+    QCOMPARE(leftSpy.first().first().value<quint32>(), m_display->serial());
+
+    QVERIFY(!keyboard->enteredSurface());
+    QVERIFY(!ckeyboard.enteredSurface());
 
     // enter it again
-    serverKeyboard->setFocusedSurface(serverSurface);
+    m_seatInterface->setFocusedKeyboardSurface(serverSurface);
     QVERIFY(modifierSpy.wait());
-    QCOMPARE(serverKeyboard->focusedSurface(), serverSurface);
+    QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
+    QCOMPARE(m_seatInterface->focusedKeyboard()->focusedSurface(), serverSurface);
+    QCOMPARE(enteredSpy.count(), 2);
+
+    QCOMPARE(keyboard->enteredSurface(), s);
+    QCOMPARE(ckeyboard.enteredSurface(), s);
 
     delete s;
     wl_display_flush(m_connection->display());
     QTest::qWait(100);
-    QVERIFY(!serverKeyboard->focusedSurface());
+    QVERIFY(!m_seatInterface->focusedKeyboardSurface());
+    QVERIFY(!m_seatInterface->focusedKeyboard());
 }
 
 void TestWaylandSeat::testCast()
@@ -565,5 +707,80 @@ void TestWaylandSeat::testDestroy()
     p->destroy();
 }
 
+void TestWaylandSeat::testSelection()
+{
+    using namespace KWayland::Client;
+    using namespace KWayland::Server;
+    QScopedPointer<DataDeviceManagerInterface> ddmi(m_display->createDataDeviceManager());
+    ddmi->create();
+    Registry registry;
+    QSignalSpy dataDeviceManagerSpy(&registry, SIGNAL(dataDeviceManagerAnnounced(quint32,quint32)));
+    QVERIFY(dataDeviceManagerSpy.isValid());
+    registry.setEventQueue(m_queue);
+    registry.create(m_connection->display());
+    QVERIFY(registry.isValid());
+    registry.setup();
+
+    QVERIFY(dataDeviceManagerSpy.wait());
+    QScopedPointer<DataDeviceManager> ddm(registry.createDataDeviceManager(dataDeviceManagerSpy.first().first().value<quint32>(),
+                                                                           dataDeviceManagerSpy.first().last().value<quint32>()));
+    QVERIFY(ddm->isValid());
+
+    QScopedPointer<DataDevice> dd1(ddm->getDataDevice(m_seat));
+    QVERIFY(dd1->isValid());
+    QSignalSpy selectionSpy(dd1.data(), SIGNAL(selectionOffered(KWayland::Client::DataOffer*)));
+    QVERIFY(selectionSpy.isValid());
+    QSignalSpy selectionClearedSpy(dd1.data(), SIGNAL(selectionCleared()));
+    QVERIFY(selectionClearedSpy.isValid());
+
+    QSignalSpy surfaceCreatedSpy(m_compositorInterface, SIGNAL(surfaceCreated(KWayland::Server::SurfaceInterface*)));
+    QVERIFY(surfaceCreatedSpy.isValid());
+    QScopedPointer<Surface> surface(m_compositor->createSurface());
+    QVERIFY(surface->isValid());
+    QVERIFY(surfaceCreatedSpy.wait());
+    auto serverSurface = surfaceCreatedSpy.first().first().value<SurfaceInterface*>();
+    m_seatInterface->setFocusedKeyboardSurface(serverSurface);
+    QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
+    QVERIFY(!m_seatInterface->focusedKeyboard());
+    serverSurface->client()->flush();
+    QCoreApplication::processEvents();
+    QCoreApplication::processEvents();
+    QVERIFY(selectionSpy.isEmpty());
+    QVERIFY(selectionClearedSpy.isEmpty());
+
+    // now let's try to set a selection - we have keyboard focus, so it should be sent to us
+    QScopedPointer<DataSource> ds(ddm->createDataSource());
+    QVERIFY(ds->isValid());
+    ds->offer(QStringLiteral("text/plain"));
+    dd1->setSelection(0, ds.data());
+    QVERIFY(selectionSpy.wait());
+    QCOMPARE(selectionSpy.count(), 1);
+    QVERIFY(selectionClearedSpy.isEmpty());
+    auto df = selectionSpy.first().first().value<DataOffer*>();
+    QCOMPARE(df->offeredMimeTypes().count(), 1);
+    QCOMPARE(df->offeredMimeTypes().first().name(), QStringLiteral("text/plain"));
+
+    // try to clear
+    dd1->setSelection(0);
+    QVERIFY(selectionClearedSpy.wait());
+    QCOMPARE(selectionClearedSpy.count(), 1);
+    QCOMPARE(selectionSpy.count(), 1);
+
+    // unset the keyboard focus
+    m_seatInterface->setFocusedKeyboardSurface(nullptr);
+    QVERIFY(!m_seatInterface->focusedKeyboardSurface());
+    QVERIFY(!m_seatInterface->focusedKeyboard());
+    serverSurface->client()->flush();
+    QCoreApplication::processEvents();
+    QCoreApplication::processEvents();
+
+    // try to set Selection
+    dd1->setSelection(0, ds.data());
+    wl_display_flush(m_connection->display());
+    QCoreApplication::processEvents();
+    QCoreApplication::processEvents();
+    QCOMPARE(selectionSpy.count(), 1);
+}
+
 QTEST_GUILESS_MAIN(TestWaylandSeat)
 #include "test_wayland_seat.moc"
